import argparse
from tempfile import mkstemp
import subprocess

from .isa import *
from . import __version__

class MachineDecodeError(Exception):
    pass

def decode(machinecode: int):
    opcode = machinecode & 0x7F
    for icls in get_insns():
        if icls._opcode == opcode and icls._match(machinecode):
            i = icls()
            i.decode(machinecode)
            return i
    raise MachineDecodeError()


def read_from_binary(fname: str):
    with open(fname, "rb") as f:
        insn = f.read(4)
        while insn:
            yield decode(int.from_bytes(insn, 'little'))
            insn = f.read(4)

def machinsn_decode():
    parser = argparse.ArgumentParser(description='Disassemble a machine instruction.')
    parser.add_argument('--version', help='Display version', action='version', version=__version__)
    subparsers = parser.add_subparsers()
    parser_cmdline = subparsers.add_parser('hexstring', help='From commandline hexstrings')
    parser_cmdline.add_argument('insn', type=str, nargs='+', help='Instruction(s) as hexstring (0x...)')
    parser_file = subparsers.add_parser('objfile', help='Read from object file')
    parser_file.add_argument('filename', type=str, help='Filename')
    parser.add_argument('--objcopy', type=str, default="riscv32-unknown-elf-objcopy", help='objcopy executable')
    args = parser.parse_args()

    if "insn" in args:
        for i in args.insn:
            print(decode(int(i,16)))
    elif "filename" in args:
        temp = mkstemp(suffix='.bin')
        subprocess.call([args.objcopy, '-O', 'binary', args.filename, temp[1]])

        for i in read_from_binary(temp[1]):
            print(i)
    else:
        parser.print_help()




class Variant(object):
    def __init__(self):
        self.intregs = None


class VariantRV32I(Variant):
    def __init__(self):
        super(VariantRV32I, self).__init__()
        self.intregs = 32


import random
import argparse
import sys
from tempfile import mkstemp
import subprocess

from .isa import *
from .variant import Variant, VariantRV32I
from .program import read_from_binary
from . import __version__


def random_instruction(variant: Variant, pool = None):
    if pool is None:
        pool = get_insns(Instruction)

    while True:
        c = random.choice(pool)
        i = c()
        i.randomize(variant)
        yield i

def random_asm(N, pool=None):
    v = VariantRV32I()
    for i in range(N):
        yield next(random_instruction(v, pool=pool))


def gen_asm(argv=None):
    parser = argparse.ArgumentParser(description='Generate sequence of assembler instructions.')
    parser.add_argument('N', nargs='?', default=10, type=int, help='Number of assembler instructions')
    parser.add_argument('-i', action='append', type=str, choices=get_mnenomics(), help='Restrict to instructions')
    parser.add_argument('--version', help='Display version', action='version', version=__version__)

    if argv is None:
        argv = sys.argv[1:]
    args = parser.parse_args(argv)

    if args.i is None:
        args.i = get_mnenomics()

    pool = [reverse_lookup(m) for m in args.i]

    for asm in random_asm(args.N, pool=pool):
        print(asm)

def check_asm_run(N, pool, compiler, objcopy):
    print("Check {} instructions from {}".format(N, [i._mnemonic for i in pool]))

    scoreboard = []

    asm = mkstemp(suffix='.S')
    objfile = mkstemp(suffix='.o')
    f = open(asm[1], "w")
    for a in random_asm(N, pool):
        scoreboard.append(a)
        f.write("{}\n".format(str(a)))
    f.close()

    subprocess.call([compiler, '-o', objfile[1], '-c', asm[1]])

    binfile = mkstemp(suffix='.bin')
    subprocess.call([objcopy, '-O', 'binary', objfile[1], binfile[1]])

    j = 0
    for i in read_from_binary(binfile[1]):
        if str(i) != str(scoreboard[j]):
            print("Check failed: {} {}".format(N, [i._mnemonic for i in pool]))
            print("{} != {}".format(i, scoreboard[j]))
            return

        j += 1
    print("Check passed: {} {}".format(N, [i._mnemonic for i in pool]))


def check_asm(argv=None):
    parser = argparse.ArgumentParser(description='Automatically test if sequences of assembler instructions compile.')
    parser.add_argument('N', nargs='?', default=100, type=int, help='Number of assembler instructions')
    parser.add_argument('-i', action='append', type=str, choices=get_mnenomics(), help='Restrict to instructions')
    parser.add_argument('-s', action='store_true', default=False, help='Test each mnemonic individually')
    parser.add_argument('--cc', type=str, default="riscv32-unknown-elf-gcc", help='Compiler executable')
    parser.add_argument('--objcopy', type=str, default="riscv32-unknown-elf-objcopy", help='objcopy executable')
    parser.add_argument('--version', help='Display version', action='version', version=__version__)

    if argv is None:
        argv = sys.argv[1:]
    args = parser.parse_args(argv)

    if args.i is None:
        args.i = get_mnenomics()

    pool = [reverse_lookup(m) for m in args.i]

    if (args.s):
        for i in pool:
            check_asm_run(args.N, [i], args.cc, args.objcopy)
    else:
        check_asm_run(args.N, pool, args.cc, args.objcopy)

if __name__ == "__main__":
    if sys.argv[1] == "random_asm":
        gen_asm(sys.argv[2:])
    elif sys.argv[1] == "check_asm":
        check_asm(sys.argv[2:])



from random import randrange, randint
from abc import ABCMeta, abstractmethod

from .variant import Variant
from .model import Model


class InvalidImmediateException(Exception):
    """
    This exception is generated by Immediates for invalid values. It contains the message for the reason.
    """
    pass

class Immediate(object):
    """
    Immediate values are stored in this container, which safeguards them. An Immediate is configured for a bit width and
    can be signed or unsigned. Finally, there are immediates in RISC-V which are aligned to instruction address
    granularity, so that an immediate can be configured to be aligned to 16-bit boundaries (lsb = 0).

    :param bits: bit width of the immediate
    :type bits: int
    :param signed: Signedness of the immediate
    :type signed: bool
    :param lsb0: Set to True if this immediate is aligned to 16-bit boundaries
    :type lsb0: bool
    """
    def __init__(self, *, bits: int, signed: bool = False, lsb0: bool = False):
        self.bits = bits
        self.signed = signed
        self.lsb0 = lsb0
        self.value = 0
        self.tcmask = 1 << (self.bits - 1)

    def max(self) -> int:
        """
        Get the maximum value this immediate can have

        :return: Maximum value of this immediate
        """
        if self.signed:
            v = (1 << (self.bits - 1)) - 1
        else:
            v = (1 << self.bits) - 1
        if self.lsb0:
            v = v - (v % 2)
        return v

    def min(self) -> int:
        """
        Get the minimum value this immediate can have

        :return: Minimum value of this immediate
        """
        if self.signed:
            return -(1 << (self.bits - 1))
        else:
            return 0

    def exception(self, msg: str) -> InvalidImmediateException:
        # Generate exception
        message = "Immediate(bits={}, signed={}, lsb0={}) {}".format(self.bits, self.signed, self.lsb0, msg)
        return InvalidImmediateException(message)

    def set(self, value: int):
        """
        Set the immediate to a value. This function checks if the value is valid and will raise an
        :class:`InvalidImmediateException` if it doesn't.

        :param value: Value to set the immediate to
        :type value: int

        :raises InvalidImmediateException: value does not match immediate
        """
        if not isinstance(value, int):
            raise self.exception("{} is not an integer".format(value))
        if self.lsb0 and self.value % 2 == 1:
            raise self.exception("{} not power of two".format(value))
        if not self.signed and value < 0:
            raise self.exception("{} cannot be negative".format(value))
        if value < self.min() or value > self.max():
            raise self.exception("{} not in allowed range {}-{}".format(value, self.min(), self.max()))

        self.value = value

    def set_from_bits(self, value: int):
        """
        Set the immediate value from machine code bits. Those are not sign extended, so it will take care of the
        proper handling.

        :param value: Value to set the immediate to
        :type value: int
        """
        if self.signed:
            value = -(value & self.tcmask) + (value & ~self.tcmask)
        self.set(value)

    def randomize(self):
        """
        Randomize this immediate to a legal value
        """
        self.value = randint(self.min(), self.max())
        if self.lsb0:
            self.value = self.value - (self.value % 2)

    def __int__(self):
        """Convert to int"""
        return self.value.__int__()

    def __str__(self):
        """Convert to string"""
        return self.value.__str__()

    def __format__(self, format_spec):
        """Apply format spec"""
        return self.value.__format__(format_spec)


class Instruction(metaclass=ABCMeta):
    """
    Base class for instructions

    This is the abstract base class for all instruction. They are derived via their instruction type.
    """

    @abstractmethod
    def randomize(self, variant: Variant):
        """
        Randomize this instruction

        This function randomizes the instance of an instruction according to the given variant.

        :param variant: RISC-V ISA variant
        :return: nothing
        """
        pass

    @abstractmethod
    def execute(self, model: Model):
        """
        Execute this instruction

        Execute the instruction on the given model

        :param model: RISC-V core model
        :return: nothing
        """
        pass

    @abstractmethod
    def decode(self, machinecode: int):
        """
        Decode a machine code and configure this instruction from it.

        :param machinecode: Machine code as 32-bit integer
        :type machinecode: int
        """
        pass

    @abstractmethod
    def __str__(self):
        """
        Generate assembler code

        Generate the assembler code for this instruction

        :return: Assembly string
        """
        pass

    def __setattr__(self, key, value):
        if key in self.__dict__ and isinstance(self.__dict__[key], Immediate):
            raise AttributeError("Instruction does not allow to overwrite immediates, use set() on them")
        super().__setattr__(key, value)

class InstructionRType(Instruction):
    """
    R-type instructions are 3-register instructions which use two source registers and write one output register.

    :param rd: Destination register
    :type rd: int
    :param rs1: Source register 1
    :type rs1: int
    :param rs2: Source register 2
    :type rs2: int
    """
    def __init__(self, rd: int = None, rs1: int = None, rs2: int = None):
        super(InstructionRType, self).__init__()
        self.rd = rd
        self.rs1 = rs1
        self.rs2 = rs2

    def randomize(self, variant: Variant):
        self.rd = randrange(0, variant.intregs)
        self.rs1 = randrange(0, variant.intregs)
        self.rs2 = randrange(0, variant.intregs)

    def decode(self, machinecode: int):
        self.rd = (machinecode >> 7) & 0x1f
        self.rs1 = (machinecode >> 15) & 0x1f
        self.rs2 = (machinecode >> 20) & 0x1f

    def __str__(self):
        return "{} x{}, x{}, x{}".format(self._mnemonic, self.rd, self.rs1, self.rs2)


class InstructionIType(Instruction):
    """
    I-type instructions are registers that use one source register and an immediate to produce a new value for the
    destination register.

    Two specialization exist for this class: :class:`InstructionILType` for load instructions and
    :class:`InstructionISType` for instructions that shift by an immediate value.

    :param rd: Destination register
    :type rd: int
    :param rs1: Source register 1
    :type rs1: int
    :param imm: 12-bit signed immediate
    :type rs2: int
    """
    def __init__(self, rd: int = None, rs1: int = None, imm: int = None):
        super(InstructionIType, self).__init__()
        self.rd = rd
        self.rs1 = rs1
        self.imm = Immediate(bits=12, signed=True)
        if imm is not None:
            self.imm.set(imm)

    def randomize(self, variant: Variant):
        self.rd = randrange(0, variant.intregs)
        self.rs1 = randrange(0, variant.intregs)
        self.imm.randomize()

    def decode(self, machinecode: int):
        self.rd = (machinecode >> 7) & 0x1f
        self.rs1 = (machinecode >> 15) & 0x1f
        self.imm.set_from_bits((machinecode >> 20) & 0xfff)

    def __str__(self):
        return "{} x{}, x{}, {}".format(self._mnemonic, self.rd, self.rs1, self.imm)


class InstructionILType(InstructionIType):
    """
    I-type instruction specialization for stores. The produce a different assembler than the base class

    :param rd: Destination register
    :type rd: int
    :param rs1: Source register 1
    :type rs1: int
    :param imm: 12-bit signed immediate
    :type rs2: int
    """
    def __str__(self):
        return "{} x{}, {}(x{})".format(self._mnemonic, self.rd, self.imm, self.rs1)

class InstructionISType(InstructionIType):
    """
    I-Type instruction specialization for shifts by immediate. The immediate differs here (5-bit unsigned).

    :param rd: Destination register
    :type rd: int
    :param rs1: Source register 1
    :type rs1: int
    :param imm: 12-bit signed immediate
    :type rs2: int
    """
    def __init__(self, rd: int = None, rs1: int = None, shamt: int = None):
        super(InstructionISType, self).__init__()
        self.rd = rd
        self.rs1 = rs1
        self.shamt = Immediate(bits=5)

    def decode(self, machinecode: int):
        self.rd = (machinecode >> 7) & 0x1f
        self.rs1 = (machinecode >> 15) & 0x1f
        self.shamt.set_from_bits((machinecode >> 20) & 0x1f)

    def randomize(self, variant: Variant):
        self.rd = randrange(0, variant.intregs)
        self.rs1 = randrange(0, variant.intregs)
        self.shamt.randomize()

    def __str__(self):
        return "{} x{}, x{}, 0x{:02x}".format(self._mnemonic, self.rd, self.rs1, self.shamt)


class InstructionSType(Instruction):
    """
    S-type instructions are used for stores. They don't have a destination register, but two source registers.

    :param rs1: Source register for base address
    :type rs1: int
    :param rs2: Source register for data
    :type rs2: int
    :param imm: Offset of store, for calculation of address relative to rs1
    :type imm: int
    """
    def __init__(self, rs1: int = None, rs2: int = None, imm: int = None):
        super(InstructionSType, self).__init__()
        self.rs1 = rs1
        self.rs2 = rs2
        self.imm = Immediate(bits=12, signed=True)
        if imm is not None:
            self.imm.set(imm)

    def randomize(self, variant: Variant):
        self.rs1 = randrange(0, variant.intregs)
        self.rs2 = randrange(0, variant.intregs)
        self.imm.randomize()

    def decode(self, machinecode: int):
        self.rs1 = (machinecode >> 15) & 0x1f
        self.rs2 = (machinecode >> 20) & 0x1f
        imm5 = (machinecode >> 7) & 0x1f
        imm7 = (machinecode >> 25) & 0x7f
        self.imm.set_from_bits((imm7 << 5) | imm5)

    def __str__(self):
        return "{} x{}, {}(x{})".format(self._mnemonic, self.rs2, self.imm, self.rs1)


class InstructionBType(Instruction):
    """
    B-type instructions encode branches. Branches have two source registers that are compared. They then change the
    program counter by the immediate value.

    :param rs1: Source 1 for comparison
    :type rs1: int
    :param rs2: Source 2 for comparison
    :type rs2: int
    :param imm: Immediate for branch destination address calculation (13-bit, signed, 16-bit aligned)
    :type imm: int
    """
    def __init__(self, rs1: int = None, rs2: int = None, imm: int = None):
        super(InstructionBType, self).__init__()
        self.rs1 = rs1
        self.rs2 = rs2
        self.imm = Immediate(bits=13, signed=True, lsb0=True)
        if imm is not None:
            self.imm.set(imm)

    def randomize(self, variant: Variant):
        self.rs1 = randrange(0, variant.intregs)
        self.rs2 = randrange(0, variant.intregs)
        self.imm.randomize()

    def decode(self, machinecode: int):
        self.rs1 = (machinecode >> 15) & 0x1f
        self.rs2 = (machinecode >> 20) & 0x1f
        imm11 = (machinecode >> 7) & 0x1
        imm1to4 = (machinecode >> 8) & 0xf
        imm5to10 = (machinecode >> 25) & 0x3f
        imm12 = (machinecode >> 31) & 0x1
        self.imm.set_from_bits((imm12 << 12) | (imm11 << 11) | (imm5to10 << 5) | (imm1to4 << 1))

    def __str__(self):
        return "{} x{}, x{}, .{:+}".format(self._mnemonic, self.rs1, self.rs2, self.imm)


class InstructionUType(Instruction):
    """
    U-type instructions are used for constant formation and set the upper bits of a register.

    :param rd: Destination register
    :type rd: int
    :param imm: Immediate (20-bit, unsigned)
    :type imm: int
    """
    def __init__(self, rd: int = None, imm: int = None):
        super(InstructionUType, self).__init__()
        self.rd = rd
        self.imm = Immediate(bits=20)
        if imm is not None:
            self.imm.set(imm)

    def randomize(self, variant: Variant):
        self.rd = randrange(0, variant.intregs)
        self.imm.randomize()

    def decode(self, machinecode: int):
        self.rd = (machinecode >> 7) & 0x1f
        self.imm.set_from_bits((machinecode >> 12) & 0xfffff)

    def __str__(self):
        return "{} x{}, {}".format(self._mnemonic, self.rd, self.imm)


class InstructionJType(Instruction):
    """
    J-type instruction are used for jump and link instructions.

    :param rd: Destination register
    :type rd: int
    :param imm: Immediate for the jump (21-bit, signed, 16-bit aligned)
    :type imm: int
    """
    def __init__(self, rd: int = None, imm: int = None):
        super(InstructionJType, self).__init__()
        self.rd = rd
        self.imm = Immediate(bits=21,signed=True,lsb0=True)
        if imm is not None:
            self.imm.set(imm)

    def randomize(self, variant: Variant):
        self.rd = randrange(0, variant.intregs)
        self.imm.randomize()

    def decode(self, machinecode: int):
        self.rd = (machinecode >> 7) & 0x1f
        imm12to19 = (machinecode >> 12) & 0xff
        imm11 = (machinecode >> 20) & 0x1
        imm1to10 = (machinecode >> 21) & 0x3ff
        imm20 = (machinecode >> 31) & 0x1
        self.imm.set_from_bits((imm20 << 20) | (imm12to19 << 12) | (imm11 << 11) | (imm1to10 << 1))

    def __str__(self):
        return "{} x{}, .{:+}".format(self._mnemonic, self.rd, self.imm)


def isa(mnemonic: str, opcode: int, funct3: int=None, funct7: int=None):
    """
    Decorator for the instructions. The decorator contains the static information for the instructions, in particular
    the encoding parameters and the assembler mnemonic.

    :param mnemonic: Assembler mnemonic
    :param opcode: Opcode of this instruction
    :param funct3: 3 bit function code on bits 14 to 12 (R-, I-, S- and B-type)
    :param funct7: 7 bit function code on bits 31 to 25 (R-type)
    :return: Wrapper class that overwrites the actual definition and contains static data
    """
    def wrapper(wrapped):
        """Get wrapper"""
        class WrappedClass(wrapped):
            """Generic wrapper class"""
            _mnemonic = mnemonic
            _opcode = opcode
            _funct3 = funct3
            _funct7 = funct7

            @staticmethod
            def _match(machinecode: int):
                """Try to match a machine code to this instruction"""
                f3 = (machinecode >> 12) & 0x7
                f7 = (machinecode >> 25) & 0x7f
                if funct3 is not None and f3 != funct3:
                    return False
                if funct7 is not None and f7 != funct7:
                    return False
                return True

        WrappedClass.__name__ = wrapped.__name__
        WrappedClass.__module__ = wrapped.__module__
        WrappedClass.__qualname__ = wrapped.__qualname__
        return WrappedClass
    return wrapper


def get_insns(cls = None):
    """
    Get all Instructions. This is based on all known subclasses of `cls`. If non is given, all Instructions are returned.
    Only such instructions are returned that can be generated, i.e., that have a mnemonic, opcode, etc. So other
    classes in the hierarchy are not matched.

    :param cls: Base class to get list
    :type cls: Instruction
    :return: List of instructions
    """
    insns = []

    if cls is None:
        cls = Instruction

    if "_mnemonic" in cls.__dict__.keys():
        insns = [cls]

    for subcls in cls.__subclasses__():
        insns += get_insns(subcls)

    return insns


def reverse_lookup(mnemonic: str):
    """
    Find instruction that matches the mnemonic.

    :param mnemonic: Mnemonic to match
    :return: :class:`Instruction` that matches or None
    """
    for i in get_insns():
        if "_mnemonic" in i.__dict__ and i._mnemonic == mnemonic:
            return i

    return None


def get_mnenomics():
    """
    Get all known mnemonics

    :return: List of all known mnemonics
    :rtype: List[str]
    """
    return [i._mnemonic for i in get_insns()]



from riscvmodel.insn import *

@isa("lui", 0x37)
class InstructionLUI(InstructionUType):
    def execute(self, model: Model):
        model.intreg[self.rd] = (model.intreg[self.rd] & 0xFFF) | (self.imm << 12)


@isa("auipc", 0x17)
class InstructionAUIPC(InstructionUType):
    def execute(self, model: Model):
        pass


@isa("jal", 0x6F)
class InstructionJAL(InstructionJType):
    def execute(self, model: Model):
        pass


@isa("jalr", 0x67, 0)
class InstructionJALR(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("beq", 0x63, 0)
class InstructionBEQ(InstructionBType):
    def execute(self, model: Model):
        if model.intreg[self.rs1] == model.intreg[self.rs2]:
            model.pc = model.pc + self.imm


@isa("bne", 0x63, 1)
class InstructionBNE(InstructionBType):
    def execute(self, model: Model):
        if model.intreg[self.rs1] != model.intreg[self.rs2]:
            model.pc = model.pc + self.imm


@isa("blt", 0x63, 4)
class InstructionBLT(InstructionBType):
    def execute(self, model: Model):
        if model.intreg[self.rs1] < model.intreg[self.rs2]:
            model.pc = model.pc + self.imm


@isa("bge", 0x63, 5)
class InstructionBGE(InstructionBType):
    def execute(self, model: Model):
        if model.intreg[self.rs1] >= model.intreg[self.rs2]:
            model.pc = model.pc + self.imm


@isa("bltu", 0x63, 6)
class InstructionBLTU(InstructionBType):
    def execute(self, model: Model):
        if model.intreg[self.rs1] < model.intreg[self.rs2]:
            model.pc = model.pc + self.imm


@isa("bgeu", 0x63, 7)
class InstructionBGEU(InstructionBType):
    def execute(self, model: Model):
        if model.intreg[self.rs1] >= model.intreg[self.rs2]:
            model.pc = model.pc + self.imm


@isa("lb", 0x03, 0)
class InstructionLB(InstructionILType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.lb(model.intreg[self.rs1] + self.imm)


@isa("lh", 0x03, 1)
class InstructionLH(InstructionILType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.lh(model.intreg[self.rs1] + self.imm)


@isa("lw", 0x03, 2)
class InstructionLW(InstructionILType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.lw(model.intreg[self.rs1] + self.imm)


@isa("lbu", 0x03, 4)
class InstructionLBU(InstructionILType):
    def execute(self, model: Model):
        pass


@isa("lhu", 0x03, 5)
class InstructionLHU(InstructionILType):
    def execute(self, model: Model):
        pass


@isa("sb", 0x23, 0)
class InstructionSB(InstructionSType):
    def execute(self, model: Model):
        pass


@isa("sh", 0x23, 1)
class InstructionSH(InstructionSType):
    def execute(self, model: Model):
        pass


@isa("sw", 0x23, 2)
class InstructionSW(InstructionSType):
    def execute(self, model: Model):
        pass


@isa("addi", 0x13, 0)
class InstructionADDI(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("slti", 0x13, 2)
class InstructionSLTI(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("sltiu", 0x13, 3)
class InstructionSLTIU(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("xori", 0x13, 4)
class InstructionXORI(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("ori", 0x13, 6)
class InstructionORI(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("andi", 0x13, 7)
class InstructionANDI(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("slli", 0x13, 1, 0x00)
class InstructionSLLI(InstructionISType):
    def execute(self, model: Model):
        pass


@isa("srli", 0x13, 5, 0x00)
class InstructionSRLI(InstructionISType):
    def execute(self, model: Model):
        pass


@isa("srai", 0x13, 5, 0x20)
class InstructionSRAI(InstructionISType):
    def execute(self, model: Model):
        pass


@isa("add", 0x33, 0, 0x00)
class InstructionADD(InstructionRType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.intreg[self.rs1] + model.intreg[self.rs2]


@isa("sub", 0x33, 0, 0x20)
class InstructionSUB(InstructionRType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.intreg[self.rs1] - model.intreg[self.rs2]


@isa("sll", 0x33, 1, 0x00)
class InstructionSLL(InstructionRType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.intreg[self.rs1] << model.intreg[self.rs2]


@isa("slt", 0x33, 2, 0x00)
class InstructionSLT(InstructionRType):
    def execute(self, model: Model):
        if model.intreg[self.rs1] < model.intreg[self.rs2]:
            model.intreg[self.rd] = 1
        else:
            model.intreg[self.rd] = 0


@isa("sltu", 0x33, 3, 0x00)
class InstructionSLTU(InstructionRType):
    def execute(self, model: Model):
        if model.intreg[self.rs1] < model.intreg[self.rs2]:
            model.intreg[self.rd] = 1
        else:
            model.intreg[self.rd] = 0


@isa("xor", 0x33, 4, 0x00)
class InstructionXOR(InstructionRType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.intreg[self.rs1] ^ model.intreg[self.rs2]


@isa("srl", 0x33, 5, 0x00)
class InstructionSRL(InstructionRType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.intreg[self.rs1] >> model.intreg[self.rs2]


@isa("sra", 0x33, 5, 0x20)
class InstructionSRA(InstructionRType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.intreg[self.rs1] >> model.intreg[self.rs2]


@isa("or", 0x33, 6, 0x00)
class InstructionOR(InstructionRType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.intreg[self.rs1] | model.intreg[self.rs2]


@isa("and", 0x33, 7, 0x00)
class InstructionAND(InstructionRType):
    def execute(self, model: Model):
        model.intreg[self.rd] = model.intreg[self.rs1] & model.intreg[self.rs2]



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="riscv-model",
    use_scm_version={
        "relative_to": __file__,
        "write_to": "riscvmodel/version.py",
    },
    author="Stefan Wallentowitz",
    author_email="stefan@wallentowitz.de",
    description="RISC-V Model",
    long_description=long_description,
    url="https://github.com/wallento/riscv-python-model",
    packages=setuptools.find_packages(),
    entry_points={
      'console_scripts': [
         'riscv-random-asm = riscvmodel.random:gen_asm',
         'riscv-random-asm-check = riscvmodel.random:check_asm',
         'riscv-machinsn-decode = riscvmodel.program:machinsn_decode'
      ],
    },
    setup_requires=[
        'setuptools_scm',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)